"""This module contains tests of the functionality of db_util."""

import unittest
import uuid
from datetime import datetime, timedelta
import time

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.database.queues import QueueStatus
from OpenOrchestrator.database.triggers import TriggerStatus

from OpenOrchestrator.tests import db_test_util


class TestDBUtil(unittest.TestCase):
    """Test functionality of db_util."""
    def setUp(self) -> None:
        db_test_util.establish_clean_database()

    def test_logs(self):
        """Test creation of logs and retrieval by different filters."""
        # Create some logs
        creation_time = datetime.now() - timedelta(seconds=2)

        for i in range(3):
            db_util.create_log(f"Test {i}", LogLevel.TRACE, None, "Message")
            db_util.create_log(f"Test {i}", LogLevel.INFO, uuid.uuid4(), "Message")
            db_util.create_log(f"Test {i}", LogLevel.ERROR, str(uuid.uuid4()), "Message")

        # Get all logs
        logs = db_util.get_logs(0, 100)
        self.assertEqual(len(logs), 9)

        # Filter by level
        logs = db_util.get_logs(0, 100, log_level=LogLevel.TRACE)
        self.assertEqual(len(logs), 3)

        # Get unique processes
        process_names = db_util.get_unique_log_process_names()
        self.assertEqual(len(process_names), 3)

        # Filter by process
        logs = db_util.get_logs(0, 100, process_name=process_names[0])
        self.assertEqual(len(logs), 3)

        # Filter by process and level
        logs = db_util.get_logs(0, 100, log_level=LogLevel.INFO, process_name=process_names[1])
        self.assertEqual(len(logs), 1)

        # Filter by date
        logs = db_util.get_logs(0, 100, from_date=creation_time)
        self.assertEqual(len(logs), 9)

        logs = db_util.get_logs(0, 100, to_date=creation_time)
        self.assertEqual(len(logs), 0)

    def test_constants(self):
        """Test all things constants."""
        # Create some constants
        db_util.create_constant("Constant1", "Value1")
        db_util.create_constant("Constant2", "Value2")

        # Get all constants
        constants = db_util.get_constants()
        self.assertEqual(len(constants), 2)

        # Get specific constant
        constant = db_util.get_constant("Constant1")
        self.assertEqual(constant.value, "Value1")
        created_at = constant.changed_at
        self.assertIsInstance(created_at, datetime)

        # Update constant
        time.sleep(0.1)
        db_util.update_constant("Constant1", "New Value")
        constant = db_util.get_constant("Constant1")
        self.assertEqual(constant.value, "New Value")
        self.assertGreater(constant.changed_at, created_at)

        # Delete constant
        db_util.delete_constant("Constant1")
        with self.assertRaises(ValueError):
            db_util.get_constant("Constant1")

    def test_credentials(self):
        """Test all things credentials."""
        # Create some credentials
        db_util.create_credential("Cred1", "User1", "Pass")
        db_util.create_credential("Cred2", "User2", "Pass")

        # Get all credentials
        creds = db_util.get_credentials()
        self.assertEqual(len(creds), 2)

        # Check encryption salt
        self.assertNotEqual(creds[0].password, creds[1].password)

        # Get credential
        cred = db_util.get_credential("Cred1")
        self.assertEqual(cred.username, "User1")
        self.assertEqual(cred.password, "Pass")
        created_at = cred.changed_at
        self.assertIsInstance(created_at, datetime)

        # Update credential
        time.sleep(0.1)
        db_util.update_credential("Cred1", "New User", "New Pass")
        cred = db_util.get_credential("Cred1")
        self.assertEqual(cred.username, "New User")
        self.assertEqual(cred.password, "New Pass")
        self.assertGreater(cred.changed_at, created_at)

        # Delete credential
        db_util.delete_credential("Cred1")
        with self.assertRaises(ValueError):
            db_util.get_credential("Cred1")

    def test_queue_elements(self):
        """Test all things queue elements."""
        # Create some queue elements
        creation_time = datetime.now() - timedelta(seconds=2)

        db_util.create_queue_element("Queue")
        db_util.create_queue_element("Queue", reference="Ref", data="Data", created_by="Me")

        # Bulk create queue elements
        refs = tuple(f"Ref{i}" for i in range(10))
        data = (None,) * 10
        db_util.bulk_create_queue_elements("Bulk", references=refs, data=data)

        with self.assertRaises(ValueError):
            db_util.bulk_create_queue_elements("Bulk", ("Ref",), ())

        # Get elements
        elements = db_util.get_queue_elements("Queue")
        self.assertEqual(len(elements), 2)

        # Get next element
        element = db_util.get_next_queue_element("Queue", set_status=False)
        self.assertIsNotNone(element)
        self.assertEqual(element.status, QueueStatus.NEW)

        element = db_util.get_next_queue_element("Queue")
        self.assertIsNotNone(element)
        self.assertEqual(element.status, QueueStatus.IN_PROGRESS)

        element2 = db_util.get_next_queue_element("Queue")
        self.assertIsNotNone(element2)
        self.assertNotEqual(element, element2)

        # Update element
        db_util.set_queue_element_status(element.id, QueueStatus.DONE, "Message")

        # Filter elements
        elements = db_util.get_queue_elements("Queue", reference="Ref")
        self.assertEqual(len(elements), 1)

        elements = db_util.get_queue_elements("Queue", status=QueueStatus.DONE)
        self.assertEqual(len(elements), 1)

        elements = db_util.get_queue_elements("Foo")
        self.assertEqual(len(elements), 0)

        elements = db_util.get_queue_elements("Queue", reference="Foo")
        self.assertEqual(len(elements), 0)

        # Filter by date
        logs = db_util.get_queue_elements("Queue", from_date=creation_time)
        self.assertEqual(len(logs), 2)

        logs = db_util.get_queue_elements("Queue", to_date=creation_time)
        self.assertEqual(len(logs), 0)

        tomorrow = datetime.now() + timedelta(days=1)
        logs = db_util.get_queue_elements("Queue", from_date=creation_time, to_date=tomorrow)
        self.assertEqual(len(logs), 2)

        # Delete element
        db_util.delete_queue_element(element.id)
        elements = db_util.get_queue_elements("Queue")
        self.assertEqual(len(elements), 1)

        # Queue count
        count = db_util.get_queue_count()
        self.assertEqual(count["Bulk"][QueueStatus.NEW], 10)

        # Get element from empty queue
        element = db_util.get_next_queue_element("Empty Queue")
        self.assertIsNone(element)

    def test_triggers(self):
        """Test generic trigger functionality."""
        db_test_util.reset_triggers()

        # Get triggers
        triggers = db_util.get_single_triggers()
        self.assertEqual(len(triggers), 1)

        triggers = db_util.get_scheduled_triggers()
        self.assertEqual(len(triggers), 1)

        triggers = db_util.get_queue_triggers()
        self.assertEqual(len(triggers), 1)

        triggers = db_util.get_all_triggers()
        self.assertEqual(len(triggers), 3)

        trigger = db_util.get_trigger(triggers[0].id)
        self.assertIsNotNone(trigger)
        self.assertEqual(trigger.id, triggers[0].id)

        # Update trigger
        trigger.process_path = "New path"
        db_util.update_trigger(trigger)
        trigger = db_util.get_trigger(trigger.id)
        self.assertEqual(trigger.process_path, "New path")

        # Delete trigger
        db_util.delete_trigger(trigger.id)
        with self.assertRaises(ValueError):
            db_util.get_trigger(trigger.id)

    def test_single_triggers(self):
        """Test running and updating single triggers."""
        db_test_util.reset_triggers()

        # Get next single trigger
        trigger = db_util.get_pending_single_triggers()[0]
        self.assertIsNotNone(trigger)

        # Begin trigger
        has_begun = db_util.begin_single_trigger(trigger.id)
        self.assertTrue(has_begun)

        # Begin already running trigger
        has_begun = db_util.begin_single_trigger(trigger.id)
        self.assertFalse(has_begun)

        # Check is running and next run
        trigger = db_util.get_trigger(trigger.id)
        self.assertEqual(trigger.process_status, TriggerStatus.RUNNING)
        self.assertIsNotNone(trigger.last_run)

        # No new trigger when the other is running
        none_trigger = db_util.get_pending_single_triggers()
        self.assertEqual(len(none_trigger), 0)

        # Set status
        db_util.set_trigger_status(trigger.id, TriggerStatus.DONE)
        trigger = db_util.get_trigger(trigger.id)
        self.assertEqual(trigger.process_status, TriggerStatus.DONE)

    def test_scheduled_triggers(self):
        """Test running and updating scheduled triggers."""
        db_test_util.reset_triggers()

        # Get next trigger
        trigger = db_util.get_pending_scheduled_triggers()[0]
        self.assertIsNotNone(trigger)

        # Begin trigger
        run_time = trigger.next_run
        has_begun = db_util.begin_scheduled_trigger(trigger.id)
        self.assertTrue(has_begun)

        # Begin already running trigger
        has_begun = db_util.begin_scheduled_trigger(trigger.id)
        self.assertFalse(has_begun)

        # Check is running- and next run
        trigger = db_util.get_trigger(trigger.id)
        self.assertGreater(trigger.next_run, run_time)
        self.assertEqual(trigger.process_status, TriggerStatus.RUNNING)
        self.assertIsNotNone(trigger.last_run)

        # No new trigger when other is running
        trigger_list = db_util.get_pending_scheduled_triggers()
        self.assertEqual(len(trigger_list), 0)

    def test_queue_triggers(self):
        """Test running and updating queue triggers."""
        db_test_util.reset_triggers()

        # Test with empty queue
        trigger_list = db_util.get_pending_queue_triggers()
        self.assertEqual(len(trigger_list), 0)

        # Test with 1 item in queue (triggers on 2)
        db_util.create_queue_element("Trigger Queue")
        trigger_list = db_util.get_pending_queue_triggers()
        self.assertEqual(len(trigger_list), 0)

        # Test with 2 items in queue
        db_util.create_queue_element("Trigger Queue")
        trigger = db_util.get_pending_queue_triggers()[0]
        self.assertIsNotNone(trigger)

        # Begin trigger
        has_begun = db_util.begin_queue_trigger(trigger.id)
        self.assertTrue(has_begun)

        # Begin already running trigger
        has_begun = db_util.begin_queue_trigger(trigger.id)
        self.assertFalse(has_begun)

        # Check is running
        trigger = db_util.get_trigger(trigger.id)
        self.assertEqual(trigger.process_status, TriggerStatus.RUNNING)
        self.assertIsNotNone(trigger.last_run)

        # No new trigger when other is running
        trigger_list = db_util.get_pending_queue_triggers()
        self.assertEqual(len(trigger_list), 0)

    def test_log_truncation(self):
        """Create logs with various lengths and test if their length is as expected"""
        # Create some logs
        long_message = "HelloWorld"*1000
        medium_message = "a"*8000
        short_message = "HelloWorld"

        db_util.create_log("TruncateTest", LogLevel.TRACE, None, long_message)
        db_util.create_log("TruncateTest", LogLevel.INFO, None, medium_message)
        db_util.create_log("TruncateTest", LogLevel.ERROR, None, short_message)

        # Test long message
        logs = db_util.get_logs(0, 100, log_level=LogLevel.TRACE)
        self.assertEqual(len(logs[0].log_message), 8000)

        # Test medium message
        logs = db_util.get_logs(0, 100, log_level=LogLevel.INFO)
        self.assertEqual(len(logs[0].log_message), len(medium_message))

        # Test short message
        logs = db_util.get_logs(0, 100, log_level=LogLevel.ERROR)
        self.assertEqual(len(logs[0].log_message), len(short_message))

    def test_schedulers(self):
        """Make pings from imitated machines and verify that they are registered."""

        db_util.send_ping_from_scheduler("Machine1")

        # Test one ping
        schedulers = db_util.get_schedulers()
        self.assertEqual(len(schedulers), 1)

        # Test multiple pings
        db_util.send_ping_from_scheduler("Machine2")
        db_util.send_ping_from_scheduler("Machine3")
        schedulers = db_util.get_schedulers()
        self.assertEqual(len(schedulers), 3)

        # Test that a ping from a machine that pinged earlier doesn't change the amount of schedulers
        db_util.send_ping_from_scheduler("Machine1")
        schedulers = db_util.get_schedulers()
        self.assertEqual(len(schedulers), 3)

        # Test types in Scheduler
        test_scheduler = schedulers[0]
        self.assertIsInstance(test_scheduler.machine_name, str)
        self.assertIsInstance(test_scheduler.last_update, datetime)

        # Test trigger
        db_util.start_trigger_from_machine("Machine1", "Test Trigger")
        schedulers = db_util.get_schedulers()
        test_scheduler = schedulers[0]
        self.assertIsInstance(test_scheduler.latest_trigger, str)
        self.assertIsInstance(test_scheduler.latest_trigger_time, datetime)


if __name__ == '__main__':
    unittest.main()



"""This module handles the connection to the database in OpenOrchestrator."""
# pylint: disable=too-many-lines

from datetime import datetime
from uuid import UUID

from cronsim import CronSim
from sqlalchemy import Engine, create_engine, select, insert, desc, text

from sqlalchemy import exc as alc_exc
from sqlalchemy import func as alc_func
from sqlalchemy.orm import Session, selectin_polymorphic

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database.logs import Log, LogLevel
from OpenOrchestrator.database.jobs import Job, JobStatus
from OpenOrchestrator.database.constants import Constant, Credential
from OpenOrchestrator.database.triggers import Trigger, SingleTrigger, ScheduledTrigger, QueueTrigger, TriggerStatus
from OpenOrchestrator.database.queues import QueueElement, QueueStatus
from OpenOrchestrator.database.schedulers import Scheduler
from OpenOrchestrator.database.truncated_string import truncate_message

_connection_engine: Engine | None = None


def connect(conn_string: str) -> bool:
    """Connects to the database using the given connection string.

    Args:
        conn_string: The connection string.

    Returns:
        bool: True if successful.
    """
    global _connection_engine  # pylint: disable=global-statement

    try:
        engine = create_engine(conn_string)
        engine.connect()
        _connection_engine = engine
        return True
    except (alc_exc.InterfaceError, alc_exc.ArgumentError, alc_exc.OperationalError):
        _connection_engine = None

    return False


def disconnect() -> None:
    """Disconnect from the database."""
    global _connection_engine  # pylint: disable=global-statement
    if _connection_engine:
        _connection_engine.dispose()
    _connection_engine = None


def check_database_revision() -> bool:
    """Check if the revision number of the connected database matches the expected revision."""
    try:
        with _get_session() as session:
            version = session.execute(text(
                """SELECT version_num FROM alembic_version"""
            )).scalar()
    except alc_exc.ProgrammingError:
        return False

    return version == "1c87ed320c78"


def _get_session() -> Session:
    """Check if theres a database connection and return a
    session to it.

    Raises:
        RuntimeError: If there's no connected database.

    Returns:
        A database session.
    """
    if not _connection_engine:
        raise RuntimeError("Not connected to database.")

    return Session(_connection_engine)


def get_conn_string() -> str:
    """Get the connection string.

    Returns:
        str: The connection string if any.
    """
    if not _connection_engine:
        raise RuntimeError("Not connected to database.")

    return str(_connection_engine.url)


def get_trigger(trigger_id: UUID | str) -> Trigger:
    """Get the trigger with the given id.

    Args:
        trigger_id: The id of the trigger.

    Returns:
        Trigger: The trigger with the given id.

    Raises:
        ValueError: If the trigger doesn't exist.
    """
    if isinstance(trigger_id, str):
        trigger_id = UUID(trigger_id)

    with _get_session() as session:
        query = (
            select(Trigger)
            .where(Trigger.id == trigger_id)
            .options(selectin_polymorphic(Trigger, (ScheduledTrigger, QueueTrigger, SingleTrigger)))
        )
        trigger = session.scalar(query)

    if not trigger:
        raise ValueError(f"No trigger with the given id: {trigger_id}")

    return trigger


def get_all_triggers() -> tuple[Trigger, ...]:
    """Get all triggers in the database.

    Returns:
        A tuple of Trigger objects.
    """
    with _get_session() as session:
        query = (
            select(Trigger)
            .options(selectin_polymorphic(Trigger, (ScheduledTrigger, QueueTrigger, SingleTrigger)))
        )
        return tuple(session.scalars(query))


def update_trigger(trigger: Trigger):
    """Updates an existing trigger in the database.

    Args:
        trigger: The trigger object with updated values.
    """
    with _get_session() as session:
        session.add(trigger)
        session.commit()
        session.refresh(trigger)


def get_scheduled_triggers() -> tuple[ScheduledTrigger, ...]:
    """Get all scheduled triggers from the database.

    Returns:
        A list of all scheduled triggers in the database.
    """
    with _get_session() as session:
        query = select(ScheduledTrigger)
        result = session.scalars(query).all()
        return tuple(result)


def get_single_triggers() -> tuple[SingleTrigger, ...]:
    """Get all single triggers from the database.

    Returns:
        A list of all single triggers in the database.
    """
    with _get_session() as session:
        query = select(SingleTrigger)
        result = session.scalars(query).all()
        return tuple(result)


def get_queue_triggers() -> tuple[QueueTrigger, ...]:
    """Get all queue triggers from the database.

    Returns:
        A list of all queue triggers in the database.
    """
    with _get_session() as session:
        query = select(QueueTrigger)
        result = session.scalars(query).all()
        return tuple(result)


def delete_trigger(trigger_id: UUID | str) -> None:
    """Delete the given trigger from the database.

    Args:
        trigger_id: The id of the trigger to delete.
    """
    if isinstance(trigger_id, str):
        trigger_id = UUID(trigger_id)

    with _get_session() as session:
        trigger = get_trigger(trigger_id)
        session.delete(trigger)
        session.commit()


# pylint: disable=too-many-positional-arguments
def get_logs(offset: int, limit: int,
             from_date: datetime | None = None, to_date: datetime | None = None,
             process_name: str | None = None, log_level: LogLevel | None = None,
             job_id: str | UUID | None = None,
             order_by: str | None = None, order_desc: bool = True,
             include_count: bool = False) -> tuple[Log, ...] | tuple[tuple[Log, ...], int]:
    """Get the logs from the database using filters and pagination.

    Args:
        offset: The index of the first log to get.
        limit: The number of logs to get.
        from_date: The datetime where the log time must be at or after. If none the filter is disabled.
        to_date: The datetime where the log time must be at or earlier. If none the filter is disabled.
        process_name: The process name to filter on. If none the filter is disabled.
        log_level: The log level to filter on. If none the filter is disabled.
        job_id: The job ID to filter on. If none the filter is disabled.
        order_by: Column to order the result by. If None, will use log_time.
        order_desc: Should result be in descending order. Defaults to True.
        include_count: Return a tuple with results as well as the total count of logs without limit applied.

    Returns:
        A list of logs matching the given filters.
    """
    if isinstance(job_id, str):
        job_id = UUID(job_id)

    def _apply_filters(_query):
        if from_date:
            _query = _query.where(Log.log_time >= from_date)
        if to_date:
            _query = _query.where(Log.log_time <= to_date)
        if process_name:
            _query = _query.where(Log.process_name == process_name)
        if log_level:
            _query = _query.where(Log.log_level == log_level)
        if job_id:
            _query = _query.where(Log.job_id == job_id)
        return _query

    with _get_session() as session:
        query = _apply_filters(select(Log))

        # Sort mapping
        if order_by:
            key = order_by.lower().replace(" ", "_")
            if key == 'level':
                sort_col = Log.log_level
            elif key == 'message':
                sort_col = Log.log_message
            elif key in ('short_job_id', 'job_id', 'full_job_id'):
                sort_col = Log.job_id
            elif key == 'process_name':
                sort_col = Log.process_name
            else:
                sort_col = Log.log_time
        else:
            sort_col = Log.log_time

        query = query.order_by(desc(sort_col) if order_desc else sort_col)
        query = query.offset(offset).select_from(Log)
        if limit > 0:
            query = query.limit(limit)

        result = session.scalars(query).all()
        logs_tuple = tuple(result)

        if include_count:
            count_query = _apply_filters(select(alc_func.count()).select_from(Log))  # pylint: disable=not-callable
            total_count = session.scalar(count_query)
            return logs_tuple, total_count

        return logs_tuple


def create_log(process_name: str, level: LogLevel, job_id: str | UUID | None, message: str) -> None:
    """Create a log in the logs table in the database.

    Args:
        process_name: The name of the process generating the log.
        level: The level of the log.
        message: The message of the log.
    """
    if isinstance(job_id, str):
        job_id = UUID(job_id)

    with _get_session() as session:
        log = Log(
            log_level = level,
            process_name = process_name,
            job_id = job_id,
            log_message = truncate_message(message)
        )
        session.add(log)
        session.commit()


def get_jobs(status: JobStatus | None = None, process_name: str | None = None) -> tuple[Job]:
    """Get jobs matching the requested status or process name.

    Args:
        status: Status of jobs, RUNNING, DONE, FAILED or KILLED. Defaults to None.
        process_name: Process name matching the jobs. Defaults to None.

    Returns:
        Tuple containing jobs matching the filters.
    """
    query = (
            select(Job)
            .order_by(desc(Job.start_time))
        )

    if status:
        query = query.where(Job.status == status)

    if process_name:
        query = query.where(Job.process_name == process_name)

    with _get_session() as session:
        result = session.scalars(query).all()
        return tuple(result)


def start_job(process_name: str, scheduler_name: str) -> Job:
    """Create a new job, using count of previous jobs and process name as ID.

    Args:
        process_name: Process name starting this job.

    Returns:
        Job id of newly created job.
    """
    with _get_session() as session:
        job = Job(
            process_name=process_name,
            scheduler_name=scheduler_name,
            status=JobStatus.RUNNING
        )
        session.add(job)
        session.commit()
        session.refresh(job)
        session.expunge(job)
        return job


def set_job_status(job_id: UUID | str, status: JobStatus):
    """Set status of job and update end_time, based on status.

    Args:
        job_id: Job ID to set status on.
        status: Status to set. Either RUNNING, DONE, FAILED or KILLED.
        Will set end_time to null if RUNNING or current time if not.

    Raises:
        ValueError: If no job is found, will raise error.
    """
    if isinstance(job_id, str):
        job_id = UUID(job_id)

    with _get_session() as session:
        job = session.get(Job, job_id)

        if not job:
            raise ValueError("No job with the given id was found.")

        job.status = status
        if status == JobStatus.RUNNING:
            job.end_time = None
        else:
            job.end_time = datetime.now()
        session.commit()


def get_job(job_id: UUID | str) -> Job:
    """Get a job by ID.

    Args:
        job_id: The ID of the job to get.

    Returns:
        The Job object.

    Raises:
        ValueError: If no job is found.
    """
    if isinstance(job_id, str):
        job_id = UUID(job_id)
    with _get_session() as session:
        job = session.get(Job, job_id)

        if not job:
            raise ValueError("No job with the given id was found.")

        session.expunge(job)
        return job


def get_unique_log_process_names() -> tuple[str, ...]:
    """Get a list of unique process names in the logs database.

    Returns:
        A list of unique process names.
    """
    query = (
        select(Log.process_name)
        .distinct()
        .order_by(Log.process_name)
    )

    with _get_session() as session:
        result = session.scalars(query).all()
        return tuple(result)


# pylint: disable=too-many-positional-arguments
def create_single_trigger(trigger_name: str, process_name: str, next_run: datetime,
                          process_path: str, process_args: str, is_git_repo: bool, is_blocking: bool,
                          priority: int, scheduler_whitelist: list[str] | None = None, git_branch: str | None = None) -> UUID:
    """Create a new single trigger in the database.

    Args:
        trigger_name: The name of the trigger.
        process_name: The process name.
        next_run: The datetime when the trigger should run.
        process_path: The path of the process.
        process_args: The argument string of the process.
        is_git_repo: If the process_path points to a git repo.
        is_blocking: If the process should be blocking.
        priority: The integer priority of the trigger.
        scheduler_whitelist: A list of names of schedulers the trigger may run on.
        git_branch: The specific git branch of the trigger.

    Returns:
        The id of the trigger that was created.
    """
    with _get_session() as session:
        trigger = SingleTrigger(
            trigger_name= trigger_name,
            process_name = process_name,
            process_path = process_path,
            process_args = process_args,
            is_git_repo = is_git_repo,
            is_blocking = is_blocking,
            next_run = next_run,
            priority=priority,
            scheduler_whitelist=scheduler_whitelist,
            git_branch=git_branch
        )
        session.add(trigger)
        session.commit()
        return trigger.id


# pylint: disable=too-many-positional-arguments
def create_scheduled_trigger(trigger_name: str, process_name: str, cron_expr: str, next_run: datetime,
                             process_path: str, process_args: str, is_git_repo: bool,
                             is_blocking: bool, priority: int, scheduler_whitelist: list[str] | None = None,
                             git_branch: str | None = None) -> UUID:
    """Create a new scheduled trigger in the database.

    Args:
        trigger_name: The name of the trigger.
        process_name: The process name.
        cron_expr: The cron expression of the trigger.
        next_run: The date to first run the trigger.
        process_path: The path of the process.
        process_args: The argument string of the process.
        is_git_repo: If the process_path points to a git repo.
        is_blocking: If the process should be blocking.
        priority: The integer priority of the trigger.
        scheduler_whitelist: A list of names of schedulers the trigger may run on.
        git_branch: The specific git branch of the trigger.

    Returns:
        The id of the trigger that was created.
    """
    with _get_session() as session:
        trigger = ScheduledTrigger(
            trigger_name= trigger_name,
            process_name = process_name,
            process_path = process_path,
            process_args = process_args,
            is_git_repo = is_git_repo,
            is_blocking = is_blocking,
            next_run = next_run,
            cron_expr = cron_expr,
            priority=priority,
            scheduler_whitelist=scheduler_whitelist,
            git_branch=git_branch
        )
        session.add(trigger)
        session.commit()
        return trigger.id


# pylint: disable=too-many-positional-arguments
def create_queue_trigger(trigger_name: str, process_name: str, queue_name: str, process_path: str,
                         process_args: str, is_git_repo: bool, is_blocking: bool,
                         min_batch_size: int, priority: int, scheduler_whitelist: list[str] | None = None,
                         git_branch: str | None = None) -> UUID:
    """Create a new queue trigger in the database.

    Args:
        trigger_name: The name of the trigger.
        process_name: The process name.
        queue_name: The name of the queue.
        process_path: The path of the process.
        process_args: The argument string of the process.
        is_git_repo: The is_git value of the process.
        is_blocking: The is_blocking value of the process.
        min_batch_size: The minimum number of queue elements before triggering.
        priority: The integer priority of the trigger.
        scheduler_whitelist: A list of names of schedulers the trigger may run on.
        git_branch: The specific git branch of the trigger.

    Returns:
        The id of the trigger that was created.
    """
    with _get_session() as session:
        trigger = QueueTrigger(
            trigger_name= trigger_name,
            process_name = process_name,
            process_path = process_path,
            process_args = process_args,
            is_git_repo = is_git_repo,
            is_blocking = is_blocking,
            queue_name = queue_name,
            min_batch_size = min_batch_size,
            priority=priority,
            scheduler_whitelist=scheduler_whitelist,
            git_branch=git_branch
        )
        session.add(trigger)
        session.commit()
        return trigger.id


def get_constant(name: str) -> Constant:
    """Get a constant from the database.

    Args:
        name: The name of the constant.

    Returns:
        Constant: The constant with the given name.

    Raises:
        ValueError: If no constant with the given name exists.
    """
    with _get_session() as session:
        constant = session.get(Constant, name)
        if constant is None:
            raise ValueError(f"No constant with name '{name}' was found.")
        return constant


def get_constants() -> tuple[Constant, ...]:
    """Get all constants in the database.

    Returns:
        tuple[Constants]: A list of constants.
    """
    with _get_session() as session:
        query = select(Constant).order_by(Constant.name)
        result = session.scalars(query).all()
        return tuple(result)


def create_constant(name: str, value: str) -> None:
    """Create a new constant in the database.

    Args:
        name: The name of the constant.
        value: The value of the constant.
    """
    with _get_session() as session:
        constant = Constant(name = name, value = value)
        session.add(constant)
        session.commit()


def update_constant(name: str, new_value: str) -> None:
    """Updates an existing constant with a new value.

    Args:
        name: The name of the constant to update.
        new_value: The new value of the constant.
    """
    with _get_session() as session:
        constant = session.get(Constant, name)

        if not constant:
            raise ValueError(f"No constant with name '{name}' was found.")

        constant.value = new_value
        session.commit()


def delete_constant(name: str) -> None:
    """Delete the constant with the given name from the database.

    Args:
        name: The name of the constant to delete.
    """
    with _get_session() as session:
        constant = session.get(Constant, name)
        session.delete(constant)
        session.commit()


def get_credential(name: str, decrypt_password: bool = True) -> Credential:
    """Get a credential from the database.
    The password of the credential is decrypted.

    Args:
        name: The name of the credential.
        decrypt_password: Whether to decrypt the credential password or not.

    Returns:
        Credential: The credential with the given name.

    Raises:
        ValueError: If no credential with the given name exists.
    """
    with _get_session() as session:
        credential = session.get(Credential, name)

    if credential is None:
        raise ValueError(f"No credential with name '{name}' was found.")

    if decrypt_password:
        credential.password = crypto_util.decrypt_string(credential.password)

    return credential


def get_credentials() -> tuple[Credential, ...]:
    """Get all credentials in the database.
    The passwords of the credentials are encrypted.

    Returns:
        tuple[Credential]: A list of credentials.
    """
    with _get_session() as session:
        query = select(Credential).order_by(Credential.name)
        result = session.scalars(query).all()
        return tuple(result)


def create_credential(name: str, username: str, password: str) -> None:
    """Create a new credential in the database.
    The password is encrypted before sending it to the database.

    Args:
        name: The name of the credential.
        username: The username of the credential.
        password: The password of the credential.
    """

    password = crypto_util.encrypt_string(password)

    with _get_session() as session:
        credential = Credential(
            name = name,
            username= username,
            password = password
        )
        session.add(credential)
        session.commit()


def update_credential(name: str, new_username: str, new_password: str) -> None:
    """Updates an existing credential with a new value.

    Args:
        name: The name of the credential to update.
        new_username: The new username of the credential.
        new_password: The new password of the credential.
    """
    new_password = crypto_util.encrypt_string(new_password)

    with _get_session() as session:
        credential = session.get(Credential, name)

        if not credential:
            raise ValueError(f"No credential with name '{name}' was found.")

        credential.username = new_username
        credential.password = new_password
        session.commit()


def delete_credential(name: str) -> None:
    """Delete the credential with the given name from the database.

    Args:
        name: The name of the credential to delete.
    """
    with _get_session() as session:
        constant = session.get(Credential, name)
        session.delete(constant)
        session.commit()


def begin_single_trigger(trigger_id: UUID | str) -> bool:
    """Set the status of a single trigger to 'running' and
    set the last run time to the current time.

    Args:
        trigger_id: The id of the trigger to begin.

    Returns:
        bool: True if the trigger was 'idle' and now 'running'.
    """
    if isinstance(trigger_id, str):
        trigger_id = UUID(trigger_id)

    with _get_session() as session:
        trigger = session.get(SingleTrigger, trigger_id)

        if not trigger:
            raise ValueError("No trigger with the given id was found.")

        if trigger.process_status != TriggerStatus.IDLE:
            return False

        trigger.process_status = TriggerStatus.RUNNING
        trigger.last_run = datetime.now()

        session.commit()
        return True


def get_pending_single_triggers() -> list[SingleTrigger]:
    """Get all single triggers that are ready to run.

    Returns:
        All single triggers ready to run if any.
    """
    with _get_session() as session:
        query = (
            select(SingleTrigger)
            .where(SingleTrigger.process_status == TriggerStatus.IDLE)
            .where(SingleTrigger.next_run <= datetime.now())
            .order_by(SingleTrigger.next_run)
        )
        return list(session.scalars(query))


def get_pending_scheduled_triggers() -> list[ScheduledTrigger]:
    """Get all scheduled triggers that are ready to run.

    Returns:
        All scheduled triggers ready to run if any.
    """
    with _get_session() as session:
        query = (
            select(ScheduledTrigger)
            .where(ScheduledTrigger.process_status == TriggerStatus.IDLE)
            .where(ScheduledTrigger.next_run <= datetime.now())
            .order_by(ScheduledTrigger.next_run)
        )
        return list(session.scalars(query))


def begin_scheduled_trigger(trigger_id: UUID | str) -> bool:
    """Set the status of a scheduled trigger to 'running',
    set the last run time to the current time,
    and set the next run time to the given datetime.

    Args:
        trigger_id: The id of the trigger to begin.
        next_run: The next datetime the trigger should run.

    Returns:
        bool: True if the trigger was 'idle' and now 'running'.
    """
    if isinstance(trigger_id, str):
        trigger_id = UUID(trigger_id)

    with _get_session() as session:
        trigger = session.get(ScheduledTrigger, trigger_id)

        if not trigger:
            raise ValueError("No trigger with the given id was found.")

        if trigger.process_status != TriggerStatus.IDLE:
            return False

        trigger.process_status = TriggerStatus.RUNNING
        trigger.last_run = datetime.now()
        trigger.next_run = next(CronSim(trigger.cron_expr, datetime.now()))

        session.commit()
        return True


def get_pending_queue_triggers() -> list[QueueTrigger]:
    """Get all queue triggers that are ready to run.

    Returns:
        All queue triggers that are ready to run if any.
    """

    with _get_session() as session:

        sub_query = (
            select(alc_func.count())  # pylint: disable=not-callable
            .where(QueueElement.queue_name == QueueTrigger.queue_name)
            .where(QueueElement.status == QueueStatus.NEW)
            .scalar_subquery()
        )

        query = (
            select(QueueTrigger)
            .where(QueueTrigger.process_status == TriggerStatus.IDLE)
            .where(sub_query >= QueueTrigger.min_batch_size)
        )
        return list(session.scalars(query))


def begin_queue_trigger(trigger_id: UUID | str) -> bool:
    """Set the status of a queue trigger to 'running' and
    set the last run time to the current time.

    Args:
        trigger_id: The id of the trigger to begin.

    Returns:
        bool: True if the trigger was 'idle' and now 'running'.
    """
    if isinstance(trigger_id, str):
        trigger_id = UUID(trigger_id)

    with _get_session() as session:
        trigger = session.get(QueueTrigger, trigger_id)

        if not trigger:
            raise ValueError("No trigger with the given id was found.")

        if trigger.process_status != TriggerStatus.IDLE:
            return False

        trigger.process_status = TriggerStatus.RUNNING
        trigger.last_run = datetime.now()

        session.commit()
        return True


def set_trigger_status(trigger_id: UUID | str, status: TriggerStatus) -> None:
    """Set the status of a trigger.

    Args:
        trigger_id: The id of the trigger.
        status: The new status of the trigger.
    """
    if isinstance(trigger_id, str):
        trigger_id = UUID(trigger_id)

    with _get_session() as session:
        trigger = session.get(Trigger, trigger_id)

        if not trigger:
            raise ValueError("No trigger with the given id was found.")

        trigger.process_status = status
        session.commit()


def create_queue_element(queue_name: str, reference: str | None = None, data: str | None = None, created_by: str | None = None) -> QueueElement:
    """Adds a queue element to the given queue.

    Args:
        queue_name: The name of the queue to add the element to.
        reference (optional): The reference of the queue element.
        data (optional): The data of the queue element.
        created_by (optional): The name of the creator of the queue element.

    Returns:
        QueueElement: The created queue element.
    """
    with _get_session() as session:
        q_element = QueueElement(
            queue_name = queue_name,
            data = data,
            reference = reference,
            created_by = created_by
        )
        session.add(q_element)
        session.commit()
        session.refresh(q_element)

    return q_element


def bulk_create_queue_elements(queue_name: str, references: tuple[str | None, ...], data: tuple[str | None, ...], created_by: str | None = None) -> None:
    """Insert multiple queue elements into a queue in an optimized manner.
    The lengths of both 'references' and 'data' must be equal to the number of elements to insert.

    Args:
        queue_name: The name of the queue to insert into.
        references: A tuple of reference strings for each queue element.
        data: A tuple of data strings for each queue element.
        created_by (Optional): The name of the creator of the queue elements.

    Raises:
        ValueError: If either 'references' or 'data' are empty, or if they are not equal in length.
    """
    if len(references) == 0:
        raise ValueError("No reference strings were given.")

    if len(data) == 0:
        raise ValueError("No data strings were given.")

    if len(references) != len(data):
        raise ValueError(f"The number of references and data strings don't match: {len(references)} != {len(data)}.")

    q_elements = (
        {
            "queue_name": queue_name,
            "reference": ref,
            "data": dat,
            "created_by": created_by
        }
        for ref, dat in zip(references, data)
    )

    with _get_session() as session:
        session.execute(insert(QueueElement), q_elements)  # type: ignore
        session.commit()


def get_next_queue_element(queue_name: str, reference: str | None = None, set_status: bool = True) -> QueueElement | None:
    """Gets the next queue element from the given queue that has the status 'new'.

    Args:
        queue_name: The name of the queue to retrieve from.
        reference (optional): The reference to filter on. If None the filter is disabled.
        set_status (optional): If true the queue element's status is set to 'in progress' and the start time is noted.

    Returns:
        QueueElement | None: The next queue element in the queue if any.
    """

    with _get_session() as session:
        query = (
            select(QueueElement)
            .where(QueueElement.queue_name == queue_name)
            .where(QueueElement.status == QueueStatus.NEW)
            .order_by(QueueElement.created_date)
            .limit(1)
        )

        if reference is not None:
            query = query.where(QueueElement.reference == reference)

        q_element = session.scalar(query)

        if q_element and set_status:
            q_element.status = QueueStatus.IN_PROGRESS
            q_element.start_date = datetime.now()
            session.commit()
            session.refresh(q_element)

        return q_element


def get_queue_elements(queue_name: str, reference: str | None = None, status: QueueStatus | None = None,
                       from_date: datetime | None = None, to_date: datetime | None = None,
                       offset: int = 0, limit: int | None = 100, search_term: str | None = None,
                       order_by: str | None = None, order_desc: bool = False, include_count: bool = False) -> tuple[QueueElement, ...] | tuple[tuple[QueueElement, ...], int]:
    """Get multiple queue elements from a queue. The elements are ordered by created_date.

    Args:
        queue_name: The queue to get elements from.
        reference (optional): The reference to filter by. If None the filter is disabled.
        status (optional): The status to filter by if any. If None the filter is disabled.
        offset (optional): The number of queue elements to skip.
        limit (optional): The number of queue elements to get.
        order_by (optional): Column to order the result by. If None, will use created_date.
        order_desc (optional): Should result be in descending order, only used with order_by.
        include_count (optional): Return a tuple with results as well as the total count of elements without limit applied.

    Returns:
        tuple[QueueElement] | tuple[tuple[QueueElement], int]: A tuple of queue elements or a tuple with a tuple of queue elements and an element count.
    """
    def _apply_filters(query):
        """Create filters for query, to allow for optional return of count.

        Args:
            query: The initial query on a queue name.

        Returns:
            The query object.
        """
        query = query.where(QueueElement.queue_name == queue_name)

        if from_date is not None:
            query = query.where(QueueElement.created_date >= from_date)
        if to_date is not None:
            query = query.where(QueueElement.created_date <= to_date)
        if reference is not None:
            query = query.where(QueueElement.reference == reference)
        if status is not None:
            query = query.where(QueueElement.status == status)
        if search_term is not None:
            query = query.where(QueueElement.reference.startswith(search_term) |
                                QueueElement.data.like(f"%{search_term}%") |
                                QueueElement.message.like(f"%{search_term}%"))
        return query

    with _get_session() as session:
        # Main query
        query = _apply_filters(select(QueueElement))

        if order_by:
            order_column = getattr(QueueElement, order_by, 'created_date')
        else:
            order_column = 'created_date'
        query = query.order_by(desc(order_column) if order_desc else order_column)

        if offset:
            query = query.offset(offset)
        if limit:
            query = query.limit(limit)

        result = session.scalars(query).all()
        elements_tuple = tuple(result)

        if include_count:
            count_query = _apply_filters(select(alc_func.count()))  # pylint: disable=not-callable
            total_count = session.scalar(count_query)
            return elements_tuple, total_count

        return elements_tuple


def get_queue_element(element_id: UUID | str) -> QueueElement:
    """Get a specific QueueElement from id.

    Args:
        element_id: ID of QueueElement to get.

    Returns:
        QueueElement with the requested ID.
    """
    if isinstance(element_id, str):
        element_id = UUID(element_id)
    with _get_session() as session:
        q_element = session.get(QueueElement, element_id)
        if not q_element:
            raise ValueError("No queue element with the given id was found.")
        return q_element


def update_queue_element(element_id: str, reference: str | None = None, status: QueueStatus | None = None, data: str | None = None, message: str | None = None,
                         created_by: str | None = None, created_date: datetime | None = None, start_date: datetime | None = None, end_date: datetime | None = None):
    """Update fields of specific QueueElement. Fields with value None will not be updated.

    Args:
        element_id: ID of QueueElement to update.
        reference: New value for reference. Defaults to None.
        status: New value for status. Defaults to None.
        data: New value for data. Defaults to None.
        message: New value for message. Defaults to None.
        created_by: New value for created_by. Defaults to None.
        created_date: New value for created_date. Defaults to None.
        start_date: New value for start_date. Defaults to None.
        end_date: New value for end_date. Defaults to None.
    """
    with _get_session() as session:
        query = select(QueueElement).where(QueueElement.id == element_id)
        q_element: QueueElement = session.scalar(query)

        if q_element:
            if reference:
                q_element.reference = reference
            if status:
                q_element.status = status
            if data:
                q_element.data = data
            if message:
                q_element.message = message
            if created_date:
                q_element.created_date = created_date
            if start_date:
                q_element.start_date = start_date
            if end_date:
                q_element.end_date = end_date
            if created_by:
                q_element.created_by = created_by
            session.commit()
            session.refresh(q_element)


def get_queue_count() -> dict[str, dict[QueueStatus, int]]:
    """Count the number of queue elements of each status for every queue.

    Returns:
        A dict for each queue with the count for each status. E.g. result[queue_name][status] => count.
    """
    with _get_session() as session:
        query = (
            select(QueueElement.queue_name, QueueElement.status, alc_func.count())  # pylint: disable=not-callable
            .group_by(QueueElement.queue_name)
            .group_by(QueueElement.status)
        )
        rows = session.execute(query)
        rows = tuple(rows)

    # Aggregate result into a dict
    result = {}
    for queue_name, status, count in rows:
        if queue_name not in result:
            result[queue_name] = {}
        result[queue_name][status] = count

    return result


def set_queue_element_status(element_id: UUID | str, status: QueueStatus, message: str | None = None) -> None:
    """Set the status of a queue element.
    If the new status is 'in progress' the start date is noted.
    If the new status is 'Done', 'Failed' or 'Abandoned' the end date is noted.

    Args:
        element_id: The id of the queue element to change status on.
        status: The new status of the queue element.
        message (Optional): The message to attach to the queue element. This overrides any existing messages.
    """
    if isinstance(element_id, str):
        element_id = UUID(element_id)

    with _get_session() as session:
        q_element = session.get(QueueElement, element_id)

        if not q_element:
            raise ValueError("No queue element with the given id was found.")

        q_element.status = status

        if message is not None:
            q_element.message = message

        match status:
            case QueueStatus.IN_PROGRESS:
                q_element.start_date = datetime.now()
            case QueueStatus.DONE | QueueStatus.FAILED | QueueStatus.ABANDONED:
                q_element.end_date = datetime.now()
            case _:
                pass

        session.commit()


def delete_queue_element(element_id: UUID | str) -> None:
    """Delete a queue element from the database.

    Args:
        element_id: The id of the queue element.
    """
    with _get_session() as session:
        q_element = session.get(QueueElement, element_id)
        session.delete(q_element)
        session.commit()


def get_schedulers() -> tuple[Scheduler, ...]:
    """Get Schedulers from the database"""
    with _get_session() as session:
        query = select(Scheduler).order_by(Scheduler.machine_name)
        result = session.scalars(query).all()
        return tuple(result)


def send_ping_from_scheduler(machine_name: str) -> None:
    """Send a ping from a running scheduler, updating the machine in the database.

    Args:
        machine_name: The machine pinging the Orchestrator
    """
    with _get_session() as session:
        scheduler = session.get(Scheduler, machine_name)

        if scheduler:
            scheduler.last_update = datetime.now()
        else:
            scheduler = Scheduler(machine_name=machine_name, last_update=datetime.now())
            session.add(scheduler)

        session.commit()


def start_trigger_from_machine(machine_name: str, trigger_name: str) -> None:
    """Start a trigger from a machine running a scheduler, updating the name and time for triggers in the database.

    Args:
        machine_name: The machine starting the trigger
        trigger_name: The trigger being started på the machine
    """
    with _get_session() as session:
        scheduler = session.get(Scheduler, machine_name)
        now = datetime.now()

        if scheduler:
            scheduler.last_update = now
            scheduler.latest_trigger = trigger_name
            scheduler.latest_trigger_time = now
        else:
            scheduler = Scheduler(machine_name=machine_name, last_update=now, latest_trigger=trigger_name, latest_trigger_time=now)
            session.add(scheduler)

        session.commit()



"""This module is responsible for the layout and functionality of the run tab
in Scheduler."""

from __future__ import annotations
from typing import TYPE_CHECKING

import tkinter
from tkinter import ttk
import sys

from sqlalchemy import exc as alc_exc

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.scheduler import runner, util
from OpenOrchestrator.database.triggers import TriggerStatus

if TYPE_CHECKING:
    from OpenOrchestrator.scheduler.application import Application


# pylint: disable-next=too-many-ancestors
class RunTab(ttk.Frame):
    """A ttk.frame object containing the functionality of the run tab in Scheduler."""
    def __init__(self, parent: ttk.Notebook, app: Application):
        super().__init__(parent)
        self.pack(fill='both', expand=True)

        self.app = app

        s = ttk.Style()
        s.configure('my.TButton', font=('Helvetica Bold', 24))
        self.button = ttk.Button(self, text="Run", command=self.button_click, style='my.TButton')
        self.button.pack()

        # Text area
        text_frame = tkinter.Frame(self)
        text_frame.pack()

        self.text_area = tkinter.Text(text_frame, state='disabled', wrap='none')

        # Redirect stdout to the text area instead of console
        sys.stdout.write = self.print_text

        # Add scroll bars to text area
        text_yscroll = ttk.Scrollbar(text_frame, orient='vertical', command=self.text_area.yview)
        text_yscroll.pack(side='right', fill='y')
        self.text_area.configure(yscrollcommand=text_yscroll.set)

        text_xscroll = ttk.Scrollbar(text_frame, orient='horizontal', command=self.text_area.xview)
        text_xscroll.pack(side='bottom', fill='x')
        self.text_area.configure(xscrollcommand=text_xscroll.set)

        self.text_area.pack()

    def button_click(self):
        """Callback for when the run/pause button is clicked."""
        if self.app.running:
            self.pause()
        else:
            self.run()

    def pause(self):
        """Stops the Scheduler and sets the app's status to 'paused'."""
        self.button.configure(text="Run")
        print('Paused... Please wait for all processes to stop before closing the application\n')
        self.app.running = False

    def run(self):
        """Starts the Scheduler and sets the app's status to 'running'."""
        if db_util.get_conn_string() is None:
            print("Can't start without a valid connection string. Go to the settings tab to configure the connection string")
            return
        if crypto_util.get_key() is None:
            print("Can't start without a valid encryption key. Go to the settings tab to configure the encryption key")
            return

        self.button.configure(text="Pause")
        print('Running...\n')
        self.app.running = True

        # Only start a new loop if it's not already running
        if self.app.tk.call('after', 'info') == '':
            self.app.after(0, loop, self.app)

    def print_text(self, text: str) -> None:
        """Appends text to the text area.
        Is used to replace the functionality of sys.stdout.write (print).

        Args:
            string: The string to append.
        """
        # Insert text at the end
        self.text_area.configure(state='normal')
        self.text_area.insert('end', text)

        # If the number of lines are above 1000 delete 10 lines from the top
        num_lines = int(self.text_area.index('end').split('.', maxsplit=1)[0])
        if num_lines > 1000:
            self.text_area.delete("1.0", "10.0")

        # Scroll to end
        self.text_area.see('end')
        self.text_area.configure(state='disabled')


def loop(app: Application) -> None:
    """The main loop function of the Scheduler.
    Checks heartbeats, check triggers, and schedules the next loop.

    Args:
        app: The Scheduler Application object.
    """
    try:
        send_ping_to_orchestrator()

        check_heartbeats(app)

        if app.running:
            check_triggers(app)

    except (alc_exc.OperationalError, alc_exc.ProgrammingError) as e:
        print(f"Couldn't connect to database. {e}")

    if len(app.running_jobs) == 0:
        print("Doing cleanup...")
        runner.clear_repo_folder()

    # Schedule next loop
    if app.running or len(app.running_jobs) > 0:
        print('Waiting 6 seconds...\n')
        app.after(6_000, loop, app)
    else:
        print("Scheduler is paused and no more processes are running.")


def check_heartbeats(app: Application) -> None:
    """Check if any running jobs are still running, failed or done.

    Args:
        app: The Scheduler Application object.
    """
    print('Checking heartbeats...')
    for job in app.running_jobs:
        if job.process.poll() is not None:
            if job.process.returncode == 0:
                print(f"Process '{job.trigger.process_name}' is done")
                runner.end_job(job)
            else:
                print(f"Process '{job.trigger.process_name}' failed. Check process log for more info.")
                runner.fail_job(job)

            app.running_jobs.remove(job)

        elif db_util.get_trigger(job.trigger.id).process_status == TriggerStatus.KILLING:
            runner.kill_job(job)
            print(f"Process '{job.trigger.process_name}' has been killed.")
            app.running_jobs.remove(job)

        else:
            print(f"Process '{job.trigger.process_name}' is still running")


def check_triggers(app: Application) -> None:
    """Checks any process is blocking
    and if not checks if any trigger should be run.

    Args:
        app: The Scheduler Application object.
    """
    # Check if process is blocking
    blocking = False
    for job in app.running_jobs:
        if job.trigger.is_blocking:
            print(f"Process '{job.trigger.process_name}' is blocking\n")
            blocking = True

    # Check triggers
    if not blocking:
        print('Checking triggers...')
        trigger = runner.poll_triggers(app)

        if trigger:
            job = runner.run_trigger(trigger)

            if job:
                app.running_jobs.append(job)


def send_ping_to_orchestrator():
    """Send a ping to the connected Orchestrator with the Scheduler application's name."""
    db_util.send_ping_from_scheduler(util.get_scheduler_name())



"""Tests relating to the trigger tab in Orchestrator."""

import unittest
from datetime import datetime, timedelta
import time

from selenium.webdriver.common.by import By

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.triggers import SingleTrigger, ScheduledTrigger, QueueTrigger, TriggerStatus
from OpenOrchestrator.tests.ui_tests import ui_util


class TestTriggerTab(unittest.TestCase):
    """Test functionality of the trigger tab ui."""
    def setUp(self) -> None:
        self.browser = ui_util.open_orchestrator()
        db_test_util.establish_clean_database()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_tab]").click()
        ui_util.refresh_ui(self.browser)

    def tearDown(self) -> None:
        self.browser.quit()

    @ui_util.screenshot_on_error
    def test_single_trigger_creation(self):
        """Test creation of a single trigger."""
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_tab_single_button]").click()

        # Fill form
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_trigger_input]").send_keys("Trigger name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_name_input]").send_keys("Process name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_time_input]").send_keys("12-11-2020 12:34:56")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_path_input]").send_keys("Process Path")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_git_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_branch_input]").send_keys("Branch1")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_args_input]").send_keys("Process args")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_priority_input]").send_keys("5")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_whitelist_input]").send_keys("Scheduler1\nScheduler2\n")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_save_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=popup_option1_button]").click()

        # Check result
        time.sleep(5)
        triggers = db_util.get_all_triggers()
        self.assertEqual(len(triggers), 1)
        trigger: SingleTrigger = triggers[0]
        self.assertIsInstance(trigger, SingleTrigger)

        self.assertEqual(trigger.trigger_name, "Trigger name")
        self.assertEqual(trigger.process_name, "Process name")
        self.assertEqual(trigger.next_run, datetime.strptime("12-11-2020 12:34:56", "%d-%m-%Y %H:%M:%S"))
        self.assertEqual(trigger.process_path, "Process Path")
        self.assertEqual(trigger.git_branch, "Branch1")
        self.assertEqual(trigger.is_git_repo, True)
        self.assertEqual(trigger.process_args, "Process args")
        self.assertEqual(trigger.is_blocking, True)
        self.assertEqual(trigger.priority, 5)
        self.assertEqual(trigger.scheduler_whitelist, ["Scheduler1", "Scheduler2"])

    @ui_util.screenshot_on_error
    def test_scheduled_trigger_creation(self):
        """Test creation of a scheduled trigger."""
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_tab_scheduled_button]").click()

        # Fill form
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_trigger_input]").send_keys("Trigger name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_name_input]").send_keys("Process name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_cron_input]").send_keys("1 1 * * *")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_path_input]").send_keys("Process Path")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_git_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_branch_input]").send_keys("Branch1")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_args_input]").send_keys("Process args")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_blocking_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_priority_input]").send_keys("5")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_whitelist_input]").send_keys("Scheduler1\nScheduler2\n")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_save_button]").click()

        # Check result
        time.sleep(2)
        triggers = db_util.get_all_triggers()
        self.assertEqual(len(triggers), 1)
        trigger: ScheduledTrigger = triggers[0]
        self.assertIsInstance(trigger, ScheduledTrigger)

        self.assertEqual(trigger.trigger_name, "Trigger name")
        self.assertEqual(trigger.process_name, "Process name")
        self.assertEqual(trigger.cron_expr, "1 1 * * *")
        self.assertEqual(trigger.process_path, "Process Path")
        self.assertEqual(trigger.git_branch, "Branch1")
        self.assertEqual(trigger.is_git_repo, True)
        self.assertEqual(trigger.process_args, "Process args")
        self.assertEqual(trigger.is_blocking, False)
        self.assertEqual(trigger.priority, 5)
        self.assertEqual(trigger.scheduler_whitelist, ["Scheduler1", "Scheduler2"])

    @ui_util.screenshot_on_error
    def test_queue_trigger_creation(self):
        """Test creation of a queue trigger."""
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_tab_queue_button]").click()

        # Fill form
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_trigger_input]").send_keys("Trigger name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_name_input]").send_keys("Process name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_queue_input]").send_keys("Queue Name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_batch_input]").send_keys("1")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_path_input]").send_keys("Process Path")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_git_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_branch_input]").send_keys("Branch1")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_args_input]").send_keys("Process args")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_blocking_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_priority_input]").send_keys("5")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_whitelist_input]").send_keys("Scheduler1\nScheduler2\n")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_save_button]").click()

        # Check result
        time.sleep(1)
        triggers = db_util.get_all_triggers()
        self.assertEqual(len(triggers), 1)
        trigger: QueueTrigger = triggers[0]
        self.assertIsInstance(trigger, QueueTrigger)

        self.assertEqual(trigger.trigger_name, "Trigger name")
        self.assertEqual(trigger.process_name, "Process name")
        self.assertEqual(trigger.queue_name, "Queue Name")
        self.assertEqual(trigger.min_batch_size, 11)
        self.assertEqual(trigger.process_path, "Process Path")
        self.assertEqual(trigger.git_branch, "Branch1")
        self.assertEqual(trigger.is_git_repo, True)
        self.assertEqual(trigger.process_args, "Process args")
        self.assertEqual(trigger.is_blocking, False)
        self.assertEqual(trigger.priority, 5)
        self.assertEqual(trigger.scheduler_whitelist, ["Scheduler1", "Scheduler2"])

    @ui_util.screenshot_on_error
    def test_trigger_table(self):
        """Test that data is shown correctly in the trigger table."""
        # Create some triggers
        yesterday = datetime.today() - timedelta(days=1)
        tomorrow = datetime.today() + timedelta(days=1)
        db_util.create_single_trigger("A Single trigger", "Single Process", yesterday, "Single path", "Single args", False, False, 0)
        db_util.create_scheduled_trigger("B Scheduled trigger", "Scheduled Process", "1 2 3 4 5", tomorrow, "Scheduled path", "Scheduled args", False, False, 0)
        db_util.create_queue_trigger("C Queue trigger", "Queue Process", "Queue Name", "Queue path", "Queue args", False, False, 25, 0)

        ui_util.refresh_ui(self.browser)

        table_data = ui_util.get_table_data(self.browser, "trigger_tab_trigger_table")

        # Check single trigger
        self.assertEqual(table_data[0][0], "A Single trigger")
        self.assertEqual(table_data[0][1], "0")
        self.assertEqual(table_data[0][2], "Single")
        self.assertEqual(table_data[0][3], "Idle")
        self.assertEqual(table_data[0][4], "Single Process")
        self.assertEqual(table_data[0][5], "Never")
        self.assertEqual(table_data[0][6], f"{datetime_util.format_datetime(yesterday)}\nOverdue")

        # Check scheduled trigger
        self.assertEqual(table_data[1][0], "B Scheduled trigger")
        self.assertEqual(table_data[1][1], "0")
        self.assertEqual(table_data[1][2], "Scheduled")
        self.assertEqual(table_data[1][3], "Idle")
        self.assertEqual(table_data[1][4], "Scheduled Process")
        self.assertEqual(table_data[1][5], "Never")
        self.assertEqual(table_data[1][6], f"{datetime_util.format_datetime(tomorrow)}")

        # Check queue trigger
        self.assertEqual(table_data[2][0], "C Queue trigger")
        self.assertEqual(table_data[2][1], "0")
        self.assertEqual(table_data[2][2], "Queue")
        self.assertEqual(table_data[2][3], "Idle")
        self.assertEqual(table_data[2][4], "Queue Process")
        self.assertEqual(table_data[2][5], "Never")
        self.assertEqual(table_data[2][6], "N/A")

    @ui_util.screenshot_on_error
    def test_delete_trigger(self):
        """Test deleting a trigger."""
        # Create a trigger
        db_util.create_queue_trigger("Queue trigger", "Queue Process", "Queue Name", "Queue path", "Queue args", False, False, 25, 0)
        ui_util.refresh_ui(self.browser)

        # Click trigger
        ui_util.click_table_row(self.browser, "trigger_tab_trigger_table", 0)

        # Delete trigger
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_delete_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=popup_option1_button]").click()

        # Check result
        time.sleep(1)
        triggers = db_util.get_all_triggers()
        self.assertEqual(len(triggers), 0)

    @ui_util.screenshot_on_error
    def test_enable_disable(self):
        """Test disabling and enabling a trigger."""
        # Create a trigger
        db_util.create_queue_trigger("Queue trigger", "Queue Process", "Queue Name", "Queue path", "Queue args", False, False, 25, 0)
        ui_util.refresh_ui(self.browser)

        # Click trigger
        ui_util.click_table_row(self.browser, "trigger_tab_trigger_table", 0)

        # Disable trigger
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_disable_button]").click()
        time.sleep(1)

        trigger = db_util.get_all_triggers()[0]
        self.assertEqual(trigger.process_status, TriggerStatus.PAUSED)

        # Enable trigger
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_enable_button]").click()
        time.sleep(1)

        trigger = db_util.get_all_triggers()[0]
        self.assertEqual(trigger.process_status, TriggerStatus.IDLE)

        # Close trigger
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_cancel_button]").click()

    @ui_util.screenshot_on_error
    def test_kill_button(self):
        """Test the kill button in the trigger popup."""
        trigger_id = db_util.create_single_trigger("Trigger Name", "Process Name", datetime.now(), "Path", "", False, False, 0, None)
        ui_util.refresh_ui(self.browser)

        # Check that kill button isn't there when idle
        ui_util.click_table_row(self.browser, "trigger_tab_trigger_table", 0)
        buttons = self.browser.find_elements(By.CSS_SELECTOR, "[auto-id=trigger_popup_kill_button]")
        self.assertEqual(len(buttons), 0)

        # Close popup
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_cancel_button]").click()

        # Now when running
        db_util.set_trigger_status(trigger_id, TriggerStatus.RUNNING)
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "trigger_tab_trigger_table", 0)
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_kill_button]").click()
        time.sleep(1)

        trigger = db_util.get_trigger(trigger_id)
        self.assertEqual(trigger.process_status, TriggerStatus.KILLING)

    @ui_util.screenshot_on_error
    def test_edit_trigger(self):
        """Test editing a trigger."""
        # Create a trigger
        db_util.create_queue_trigger("Queue trigger", "Queue Process", "Queue Name", "Queue path", "Queue args", False, False, 25, 0, ["Scheduler 1"], "Branch1")
        ui_util.refresh_ui(self.browser)

        # Click trigger
        ui_util.click_table_row(self.browser, "trigger_tab_trigger_table", 0)

        # Edit trigger
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_trigger_input]").send_keys(" Edit")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_name_input]").send_keys(" Edit")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_queue_input]").send_keys(" Edit")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_batch_input]").send_keys("1")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_path_input]").send_keys(" Edit")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_git_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_branch_input]").send_keys(" Edit")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_args_input]").send_keys(" Edit")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_blocking_check]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_priority_input]").send_keys("1")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_whitelist_input]").send_keys("Scheduler2\n")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=trigger_popup_save_button]").click()
        time.sleep(1)

        # Check result
        trigger: QueueTrigger = db_util.get_all_triggers()[0]
        self.assertEqual(trigger.trigger_name, "Queue trigger Edit")
        self.assertEqual(trigger.process_name, "Queue Process Edit")
        self.assertEqual(trigger.queue_name, "Queue Name Edit")
        self.assertEqual(trigger.process_path, "Queue path Edit")
        self.assertEqual(trigger.process_args, "Queue args Edit")
        self.assertEqual(trigger.is_git_repo, True)
        self.assertEqual(trigger.is_blocking, True)
        self.assertEqual(trigger.scheduler_whitelist, ["Scheduler 1", "Scheduler2"])
        self.assertEqual(trigger.git_branch, "Branch1 Edit")
        self.assertEqual(trigger.priority, 1)
        self.assertEqual(trigger.min_batch_size, 251)


if __name__ == '__main__':
    unittest.main(failfast=True)



"""Tests relating to the schedulers tab in Orchestrator."""

import unittest

from selenium.webdriver.common.by import By

from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.tests.ui_tests import ui_util


class TestSchedulersTab(unittest.TestCase):
    """Test functionality of the schedulers tab ui."""
    @classmethod
    def setUpClass(cls) -> None:
        cls.browser = ui_util.open_orchestrator()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.browser.quit()

    def setUp(self) -> None:
        db_test_util.establish_clean_database()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=schedulers_tab]").click()
        ui_util.refresh_ui(self.browser)

    @ui_util.screenshot_on_error
    def test_schedulers_table(self):
        """Test that schedulers are shown correctly in the schedulers table."""
        db_util.send_ping_from_scheduler("Testmachine")
        db_util.start_trigger_from_machine("Testmachine2", "RPA Process")
        ui_util.refresh_ui(self.browser)

        # Check that the text content is displayed correctly
        table_data = ui_util.get_table_data(self.browser, "schedulers_tab_schedulers_table")
        self.assertEqual(table_data[0][0], "Testmachine")
        self.assertRegex(table_data[0][1], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(table_data[0][2], "None yet")

        self.assertEqual(table_data[1][0], "Testmachine2")
        self.assertRegex(table_data[1][1], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(table_data[1][2], "RPA Process")
        self.assertRegex(table_data[1][3], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")


if __name__ == '__main__':
    unittest.main()



"""This module contains helper functions for automated tests."""

from nicegui import ui


def set_automation_ids(container, prefix: str):
    """Set automation ids for automated tests.
    Iterate though all attributes of the container and set 'auto-id'
    on any ui-elements.

    Args:
        container: The object that contains a number of ui-elements.
        prefix: The prefix to add to the automation ids for all elements.
    """
    for name, obj in container.__dict__.items():
        if isinstance(obj, (ui.button, ui.input, ui.checkbox, ui.number, ui.table, ui.tab, ui.select, ui.input_chips, ui.label, ui.code)):
            obj.props(f"auto-id={prefix}_{name}")



"""This module provides an input element for entering a datetime."""

from datetime import datetime
from typing import Optional, Callable, Any

from nicegui import ui


class DatetimeInput(ui.input):
    """A datetime input with a button to show a date and time picker dialog."""
    PY_FORMAT = "%d-%m-%Y %H:%M:%S"
    VUE_FORMAT = "DD-MM-YYYY HH:mm:ss"

    def __init__(self, label: str, on_change: Optional[Callable[..., Any]] = None, allow_empty: bool = False) -> None:
        """Create a new DatetimeInput.

        Args:
            label: The label for the input element.
            on_change: A callable to execute on change. Defaults to None.
            allow_empty: Whether to allow an empty input on validation. Defaults to False.
        """
        super().__init__(label, on_change=lambda: self._on_change(on_change))
        self.props("clearable")
        self.complete = False

        # Define dialog
        with ui.dialog() as self._dialog, ui.card():
            date_input = ui.date(mask=self.VUE_FORMAT).props("today-btn first-day-of-week=1")
            time_input = ui.time(mask=self.VUE_FORMAT).props("format24h")

        # Define input
        with self:
            ui.button(icon="event", on_click=self._dialog.open).props("flat")
            self.on("click", self._dialog.open)

        # Bind inputs together
        self.bind_value(date_input)
        self.bind_value(time_input)

        self._define_validation(allow_empty)

        self.complete = True

    def _define_validation(self, allow_empty: bool):
        if not allow_empty:
            self._validation = {  # pylint: disable=protected-access
                "Please enter a datetime": bool,
                f"Invalid datetime: {self.PY_FORMAT}": lambda v: self.get_datetime() is not None
            }

        else:
            def validate(value: str):
                if not value:
                    return True

                return self.get_datetime() is not None

            self._validation = {f"Invalid datetime: {self.PY_FORMAT}": validate}  # pylint: disable=protected-access

    def get_datetime(self) -> datetime | None:
        """Get the text from the input as a datetime object, if
        the current text in the input is valid else None.

        Returns:
            datetime: The value as a datetime object if any.
        """
        try:
            return datetime.strptime(self.value, self.PY_FORMAT)
        except (TypeError, ValueError):
            return None

    def set_datetime(self, value: datetime) -> None:
        """Set the value of the datetime input.

        Args:
            value: The new datetime value.
        """
        if not value:
            return
        self.value = value.strftime(self.PY_FORMAT)

    def _on_change(self, func: Optional[Callable[..., Any]]) -> None:
        """Wrapper for the input's on_change function.
        This avoids that the on_change function is called
        before the element is fully initialized.

        Args:
            func: The on_change function of the element.
        """
        if self.complete and func:
            func()



"""This module contains tests for OrchestratorConnection."""

import unittest
from datetime import datetime
from uuid import UUID, uuid4
import os

from OpenOrchestrator.orchestrator_connection.connection import OrchestratorConnection
from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.database.queues import QueueStatus

from OpenOrchestrator.tests import db_test_util


class TestOrchestratorConnection(unittest.TestCase):
    """Tests for OrchestratorConnection."""
    @classmethod
    def setUpClass(cls) -> None:
        db_test_util.establish_clean_database()
        process_name = "Process"
        cls.trigger_id = db_util.create_single_trigger(trigger_name=process_name,
                                                       process_name=process_name,
                                                       next_run=datetime.now(),
                                                       process_path="",
                                                       process_args="",
                                                       is_git_repo=False,
                                                       is_blocking=False,
                                                       priority=0
                                                       )
        cls.connection = OrchestratorConnection("Process", os.environ["CONN_STRING"], crypto_util.get_key(), "Args", str(cls.trigger_id), str(uuid4()))

    def test_trigger_pause(self):
        """Test pausing triggers."""
        self.assertFalse(self.connection.is_trigger_pausing())
        self.connection.pause_my_trigger()
        self.assertTrue(self.connection.is_trigger_pausing())

    def test_logging(self):
        """Test all three logging functions."""
        # Create some logs
        self.connection.log_trace(LogLevel.TRACE.value)
        self.connection.log_info(LogLevel.INFO.value)
        self.connection.log_error(LogLevel.ERROR.value)

        # Do the same test for each level
        for level in LogLevel:
            log = db_util.get_logs(0, 1, log_level=level)[0]
            self.assertIsNotNone(log)

            self.assertIsInstance(log.id, UUID)
            self.assertEqual(log.log_level, level)
            self.assertEqual(log.log_message, level.value)
            self.assertIsInstance(log.log_time, datetime)
            self.assertEqual(log.process_name, "Process")

        logs = db_util.get_logs(0, 100)
        self.assertEqual(len(logs), 3)

    def test_constants(self):
        """Test all things constants."""
        # Create a constant
        db_util.create_constant("Constant", "Value")

        # Get constant
        constant = self.connection.get_constant("Constant")
        self.assertIsNotNone(constant)
        self.assertEqual(constant.name, "Constant")
        self.assertEqual(constant.value, "Value")
        self.assertIsInstance(constant.changed_at, datetime)

        # Update constant
        self.connection.update_constant("Constant", "New Value")
        constant = self.connection.get_constant("Constant")
        self.assertEqual(constant.value, "New Value")

    def test_credentials(self):
        """Test all things credentials."""
        # Create credential
        db_util.create_credential("Credential", "Username", "Password")

        # Get credential
        credential = self.connection.get_credential("Credential")
        self.assertIsNotNone(credential)
        self.assertEqual(credential.name, "Credential")
        self.assertEqual(credential.username, "Username")
        self.assertEqual(credential.password, "Password")
        self.assertIsInstance(credential.changed_at, datetime)

        # Update credential
        self.connection.update_credential("Credential", "New Username", "New Password")
        credential = self.connection.get_credential("Credential")
        self.assertEqual(credential.username, "New Username")
        self.assertEqual(credential.password, "New Password")

    def test_queue_elements(self):
        """Test all things queue elements."""
        # Create elements
        self.connection.create_queue_element("Queue")
        self.connection.create_queue_element("Queue", reference="Ref", data="data", created_by="Me")

        self.connection.bulk_create_queue_elements(
            "Bulk Queue",
            references=(None,)*10,
            data=("data",)*10,
            created_by="Me"
        )

        # Get elements
        elements = self.connection.get_queue_elements("Queue")
        self.assertEqual(len(elements), 2)

        elements = self.connection.get_queue_elements("Queue", "Ref")
        self.assertEqual(len(elements), 1)

        # Get next
        element = self.connection.get_next_queue_element("Bulk Queue", set_status=False)
        self.assertIsNotNone(element)
        self.assertEqual(element.status, QueueStatus.NEW)

        element = self.connection.get_next_queue_element("Bulk Queue")
        self.assertIsNotNone(element)
        self.assertEqual(element.status, QueueStatus.IN_PROGRESS)

        element2 = self.connection.get_next_queue_element("Bulk Queue")
        self.assertNotEqual(element, element2)

        # Set status
        self.connection.set_queue_element_status(element.id, QueueStatus.DONE)
        elements = self.connection.get_queue_elements("Bulk Queue", status=QueueStatus.DONE)
        self.assertEqual(len(elements), 1)

        # Delete element
        self.connection.delete_queue_element(element.id)
        elements = self.connection.get_queue_elements("Bulk Queue")
        self.assertEqual(len(elements), 9)


if __name__ == '__main__':
    unittest.main()



"""This module is responsible for the layout and functionality of the settings tab
in Scheduler."""

import os
import tkinter
from tkinter import ttk, messagebox

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.scheduler import util


# pylint: disable=too-many-ancestors
class SettingsTab(ttk.Frame):
    """A ttk.Frame object containing the functionality of the settings tab in Scheduler."""
    def __init__(self, parent: ttk.Notebook):
        super().__init__(parent)
        self.pack(fill='both', expand=True)

        self.columnconfigure(1, weight=1)

        ttk.Label(self, text="Connection string:").grid(row=0, column=0, sticky='w')

        self._conn_entry = ttk.Entry(self)
        self._conn_entry.grid(row=0, column=1, sticky='ew')

        self._conn_button = ttk.Button(self, text="Connect", command=self._connect)
        self._conn_button.grid(row=0, column=2, sticky='e')

        ttk.Label(self, text="Encryption key:").grid(row=1, column=0, sticky='w')

        self._key_entry = ttk.Entry(self)
        self._key_entry.grid(row=1, column=1, sticky='ew')

        self._disconn_button = ttk.Button(self, text="Disconnect", command=self._disconnect, state='disabled')
        self._disconn_button.grid(row=1, column=2, sticky='e')

        ttk.Label(self, text=f"Scheduler name: {util.get_scheduler_name()}").grid(row=2, column=0, sticky='w', columnspan=2, pady=10)

        self.whitelist_value = tkinter.BooleanVar()
        self._whitelist_check = ttk.Checkbutton(self, text="Only run whitelisted triggers", variable=self.whitelist_value)
        self._whitelist_check.grid(row=3, column=0, sticky='w', columnspan=2)

        self._auto_fill()

    def _connect(self) -> None:
        """Validate the connection string and encryption key
        and connect to the database.
        """
        conn_string = self._conn_entry.get()
        crypto_key = self._key_entry.get()

        if not crypto_util.validate_key(crypto_key):
            messagebox.showerror("Invalid encryption key", "The entered encryption key is not a valid AES key.")
            return

        if db_util.connect(conn_string):
            crypto_util.set_key(crypto_key)
            self._set_state(True)
            if not db_util.check_database_revision():
                messagebox.showerror("Warning", "This version of Scheduler doesn't match the version of the connected database. Unexpected errors might occur.")

    def _disconnect(self) -> None:
        db_util.disconnect()
        crypto_util.set_key(None)
        self._set_state(False)

    def _set_state(self, connected: bool) -> None:
        if connected:
            self._conn_entry.configure(state='disabled')
            self._key_entry.configure(state='disabled')
            self._conn_button.configure(state='disabled')
            self._disconn_button.configure(state='normal')
            self._whitelist_check.configure(state='disabled')
        else:
            self._conn_entry.configure(state='normal')
            self._key_entry.configure(state='normal')
            self._conn_button.configure(state='normal')
            self._disconn_button.configure(state='disabled')
            self._whitelist_check.configure(state='normal')

    def _auto_fill(self) -> None:
        """Check the environment for a connection string
        and encryption key and autofill the inputs.
        """
        conn_string = os.environ.get('OpenOrchestratorConnString', None)
        if conn_string:
            self._conn_entry.insert(0, conn_string)

        crypto_key = os.environ.get('OpenOrchestratorKey', None)
        if crypto_key:
            self._key_entry.insert(0, crypto_key)

    def new_key(self):
        """Creates a new encryption key and inserts it
        into the key entry.
        """
        key = crypto_util.generate_key().decode()
        self._key_entry.delete(0, 'end')
        self._key_entry.insert(0, key)



"""This module is responsible for the layout and functionality of the Settings tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.common.connection_frame import ConnectionFrame
from OpenOrchestrator.orchestrator import test_helper


# pylint: disable-next=too-few-public-methods
class SettingsTab():
    """The settings tab object for Orchestrator."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            conn_frame = ConnectionFrame()
            with ui.row().classes("w-full"):
                self.key_button = ui.button("Generate Key", on_click=conn_frame.new_key)

        test_helper.set_automation_ids(self, "settings_tab")



"""This module is responsible for checking triggers and running processes."""

from __future__ import annotations
from typing import TYPE_CHECKING
import os
import shutil
import subprocess
from dataclasses import dataclass
import uuid

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.triggers import Trigger, SingleTrigger, ScheduledTrigger, QueueTrigger, TriggerStatus
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.scheduler import util
from OpenOrchestrator.database.jobs import Job, JobStatus

if TYPE_CHECKING:
    from OpenOrchestrator.scheduler.application import Application


@dataclass
class SchedulerJob():
    """An object that holds information about a running job."""
    process: subprocess.Popen
    trigger: Trigger
    job: Job
    process_folder: str | None


def poll_triggers(app: Application) -> Trigger | None:
    """Checks for pending triggers and returns the first viable one.
    This takes priority and whitelist of the triggers into account.

    Args:
        app: The Application object of the Scheduler app.

    Returns:
        The first viable trigger to run if any.
    """
    # Get all pending triggers
    trigger_list = (
        db_util.get_pending_single_triggers() +
        db_util.get_pending_scheduled_triggers() +
        db_util.get_pending_queue_triggers()
    )

    # Sort by priority and type
    # Single > Scheduled > Queue
    trigger_order = {SingleTrigger: 0, ScheduledTrigger: 1, QueueTrigger: 2}
    trigger_list.sort(key=lambda t: (-t.priority, trigger_order[type(t)]))

    # Run the first eligible trigger
    other_jobs_running = len(app.running_jobs) > 0
    scheduler_name = util.get_scheduler_name()
    is_exclusive = app.settings_tab_.whitelist_value.get()

    for trigger in trigger_list:
        whitelisted = trigger.scheduler_whitelist and scheduler_name in trigger.scheduler_whitelist  # Is the Scheduler in the whitelist
        unlisted_allowed = not is_exclusive and not trigger.scheduler_whitelist  # Is the scheduler allowed to run non-designated triggers

        if whitelisted or unlisted_allowed:
            if not (trigger.is_blocking and other_jobs_running):
                return trigger

    return None


def run_trigger(trigger: Trigger) -> SchedulerJob | None:
    """Mark a trigger as running in the database
    and start the process.

    Args:
        trigger: The trigger to run.

    Returns:
        A Job object describing the process if successful.
    """
    print('Running trigger: ', trigger.trigger_name)

    if isinstance(trigger, SingleTrigger) and db_util.begin_single_trigger(trigger.id):
        return run_process(trigger)

    if isinstance(trigger, ScheduledTrigger) and db_util.begin_scheduled_trigger(trigger.id):
        return run_process(trigger)

    if isinstance(trigger, QueueTrigger) and db_util.begin_queue_trigger(trigger.id):
        return run_process(trigger)

    return None


def clone_git_repo(repo_url: str, branch: str | None) -> str:
    """Clone the git repo at the path to %USER%\\desktop\\Scheduler_Repos\\%UUID%.

    Args:
        repo_url: URL to the git repo to clone.
        branch: The specific git branch to clone.

    Returns:
        str: The path to the cloned repo on the desktop.
    """
    repo_folder = get_repo_folder_path()
    unique_id = str(uuid.uuid4())
    repo_path = os.path.join(repo_folder, unique_id)

    os.makedirs(repo_path)

    if shutil.which('git') is None:
        raise RuntimeError('git is not installed or not found in the system PATH.')

    args = ['git', 'clone']
    if branch and branch.strip():
        args.extend(["-b", branch.strip()])
    args.extend([repo_url, repo_path])

    try:
        subprocess.run(args, check=True)
    except subprocess.CalledProcessError as exc:
        raise ValueError(f"Failed to clone git branch '{branch if branch else 'DEFAULT'}' at '{repo_url}'.") from exc

    return repo_path


def clear_repo_folder() -> None:
    """Completely remove the repos folder."""
    repo_folder = get_repo_folder_path()
    clear_folder(repo_folder)


def get_repo_folder_path() -> str:
    """Gets the path to the folder where robot repos should be saved.

    Returns:
        str: The absolute path to the repo folder.
    """
    user_path = os.path.expanduser("~")
    repo_path = os.path.join(user_path, "Desktop", "Scheduler_Repos")
    return repo_path


def clear_folder(folder_path: str) -> None:
    """Clear a folder on the system.

    Args:
        folder_path: The folder to remove.
    """
    subprocess.run(['rmdir', '/s', '/q', folder_path], check=False, shell=True, capture_output=True)


def find_main_file(folder_path: str) -> str:
    """Finds the file in the given folder with the name 'main.py'.
    The search checks subfolders recursively.
    Only the first found file is returned.

    Args:
        folder_path: The path to the folder.

    Returns:
        str: The path to the main.py file.
    """
    for dir_path, _, file_names in os.walk(folder_path):
        for file_name in file_names:
            if file_name == 'main.py':
                return os.path.join(dir_path, file_name)

    raise ValueError("No 'main.py' file found in the folder or its subfolders.")


def end_job(job: SchedulerJob) -> None:
    """Mark a job as ended in the triggers table
    in the database.
    If it's a single trigger it's marked as 'Done'
    else it's marked as 'Idle' or 'Paused'.

    Args:
        job: The job whose trigger to mark as ended.
    """
    if isinstance(job.trigger, SingleTrigger):
        db_util.set_trigger_status(job.trigger.id, TriggerStatus.DONE)

    elif isinstance(job.trigger, (ScheduledTrigger, QueueTrigger)):
        current_status = db_util.get_trigger(job.trigger.id).process_status
        if current_status == TriggerStatus.PAUSING:
            db_util.set_trigger_status(job.trigger.id, TriggerStatus.PAUSED)
        elif current_status == TriggerStatus.RUNNING:
            db_util.set_trigger_status(job.trigger.id, TriggerStatus.IDLE)
    db_util.set_job_status(job.job.id, JobStatus.DONE)

    if job.process_folder:
        clear_folder(job.process_folder)


def fail_job(job: SchedulerJob) -> None:
    """Mark a job as failed in the triggers table in the database.

    Args:
        job: The job whose trigger to mark as failed.
    """
    db_util.set_trigger_status(job.trigger.id, TriggerStatus.FAILED)
    _, error = job.process.communicate()
    error_msg = f"An uncaught error ocurred during the process:\n{error}"
    db_util.create_log(job.trigger.process_name, LogLevel.ERROR, job.job.id, error_msg)
    db_util.set_job_status(job.job.id, JobStatus.FAILED)

    if job.process_folder:
        clear_folder(job.process_folder)


def kill_job(job: SchedulerJob) -> None:
    """Kill the job's process and mark is as killed in the database.

    Args:
        job: The job whose process to kill.
    """
    # Kill process tree to ensure child processes (like browsers or subprocesses) are also killed
    # WARNING: This will cause an error on non-Windows systems
    subprocess.run(['taskkill', '/F', '/T', '/PID', str(job.process.pid)],
                   check=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    job.process.kill()
    db_util.set_trigger_status(job.trigger.id, TriggerStatus.KILLED)
    db_util.set_job_status(job.job.id, JobStatus.KILLED)

    if job.process_folder:
        clear_folder(job.process_folder)


def run_process(trigger: Trigger) -> SchedulerJob | None:
    """Runs the process of the given trigger with the necessary inputs:
    Process name
    Connection string
    Crypto key
    Process args

    If the trigger's process_path is pointing to a git repo the repo is cloned
    and the main.py file in the repo is found and run.

    Supports only .py files.

    If any exceptions occur during launch the trigger will be marked as failed.

    Args:
        trigger: The trigger whose process to run.

    Returns:
        Job: A Job object referencing the process if successful.
    """
    process_path = trigger.process_path
    folder_path = None

    machine_name = util.get_scheduler_name()
    job = db_util.start_job(trigger.process_name, machine_name)

    try:
        if trigger.is_git_repo:
            folder_path = clone_git_repo(process_path, trigger.git_branch)
            process_path = find_main_file(folder_path)

        if not os.path.isfile(process_path):
            raise ValueError(f"The process path didn't point to a file on the system. Path: '{process_path}'")

        if not process_path.endswith(".py"):
            raise ValueError(f"The process path didn't point to a valid file. Supported files are [.py]. Path: '{process_path}'")

        conn_string = db_util.get_conn_string()
        crypto_key = crypto_util.get_key()

        command_args = ['python', process_path, trigger.process_name, conn_string, crypto_key, trigger.process_args, str(trigger.id), str(job.id)]

        process = subprocess.Popen(command_args, stderr=subprocess.PIPE, text=True)  # pylint: disable=consider-using-with

        db_util.start_trigger_from_machine(machine_name, str(trigger.trigger_name))

        return SchedulerJob(process, trigger, job, folder_path)

    # We actually want to catch any exception here
    # pylint: disable=broad-exception-caught
    except Exception as exc:
        db_util.set_trigger_status(trigger.id, TriggerStatus.FAILED)
        error_msg = f"Scheduler couldn't launch the process:\n{exc.__class__.__name__}:\n{exc}"
        db_util.create_log(trigger.process_name, LogLevel.ERROR, None, error_msg)
        db_util.set_job_status(job.id, JobStatus.FAILED)
        print(error_msg)

    return None



"""Entry point for any Alembic commands."""

from logging.config import fileConfig
from sqlalchemy import engine_from_config
from sqlalchemy import pool
from alembic import context

from OpenOrchestrator.database import base, logs, triggers, queues, constants, schedulers, jobs   # noqa: F401 pylint: disable=unused-import

config = context.config

if config.config_file_name is not None:
    fileConfig(config.config_file_name)

target_metadata = base.Base.metadata

if not config.get_main_option("sqlalchemy.url"):
    x_args = context.get_x_argument()
    if len(x_args) < 1:
        raise ValueError("Please provide a connection string in the -x argument.")
    config.set_main_option("sqlalchemy.url", context.get_x_argument()[0])


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section, {}),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        # Check if we're using MSSQL or SQLite
        dialect_name = connection.dialect.name

        configure_args = {
            "connection": connection,
            "target_metadata": target_metadata,
            "include_schemas": True
        }

        # Only set version_table_schema for MSSQL
        if dialect_name == 'mssql':
            configure_args["version_table_schema"] = 'dbo'

        context.configure(**configure_args)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# test_jobs.py
"""Tests for the Job ORM and related database functions."""

import unittest
from datetime import datetime
import time

from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.jobs import JobStatus


class TestJobORM(unittest.TestCase):
    """Test the Job ORM class and database operations."""

    def setUp(self):
        db_test_util.establish_clean_database()

    def test_create_job(self):
        """Test creating a job in the database."""
        job = db_util.start_job("Test Process", "Test Scheduler")

        self.assertIsNotNone(job)
        self.assertIsNotNone(job.id)
        self.assertEqual(job.process_name, "Test Process")
        self.assertEqual(job.scheduler_name, "Test Scheduler")
        self.assertEqual(job.status, JobStatus.RUNNING)
        self.assertIsNotNone(job.start_time)
        self.assertIsNone(job.end_time)

    def test_job_default_values(self):
        """Test that default values are set correctly."""
        job = db_util.start_job("Process", "Scheduler")

        # Status should default to RUNNING
        self.assertEqual(job.status, JobStatus.RUNNING)

        # start_time should be set automatically
        self.assertIsInstance(job.start_time, datetime)
        time_diff = abs((datetime.now() - job.start_time).total_seconds())
        self.assertLess(time_diff, 1)  # Should be within 1 second

        # end_time should be None initially
        self.assertIsNone(job.end_time)

    def test_update_job_status(self):
        """Test updating job status through lifecycle."""
        job = db_util.start_job("Process", "Scheduler")
        original_id = job.id

        # Update to DONE - should set end_time automatically
        db_util.set_job_status(original_id, JobStatus.DONE)
        updated_job = db_util.get_job(original_id)
        self.assertEqual(updated_job.status, JobStatus.DONE)
        self.assertIsNotNone(updated_job.end_time)

        # Create new job and update to FAILED
        job2 = db_util.start_job("Process2", "Scheduler")
        db_util.set_job_status(job2.id, JobStatus.FAILED)
        updated_job2 = db_util.get_job(job2.id)
        self.assertEqual(updated_job2.status, JobStatus.FAILED)
        self.assertIsNotNone(updated_job2.end_time)

        # Create new job and update to KILLED
        job3 = db_util.start_job("Process3", "Scheduler")
        db_util.set_job_status(job3.id, JobStatus.KILLED)
        updated_job3 = db_util.get_job(job3.id)
        self.assertEqual(updated_job3.status, JobStatus.KILLED)
        self.assertIsNotNone(updated_job3.end_time)

    def test_end_time_set_on_completion(self):
        """Test that end_time is set when job status changes from RUNNING."""
        job = db_util.start_job("Process", "Scheduler")
        self.assertIsNone(job.end_time)

        # Complete the job
        time.sleep(0.1)
        db_util.set_job_status(job.id, JobStatus.DONE)

        updated_job = db_util.get_job(job.id)
        self.assertIsNotNone(updated_job.end_time)
        self.assertGreater(updated_job.end_time, updated_job.start_time)

    def test_get_jobs(self):
        """Test retrieving multiple jobs."""
        # Create multiple jobs
        job1 = db_util.start_job("Process1", "Scheduler1")
        time.sleep(0.01)
        job2 = db_util.start_job("Process2", "Scheduler2")
        time.sleep(0.01)
        job3 = db_util.start_job("Process3", "Scheduler1")

        # Get all jobs
        jobs = db_util.get_jobs()
        self.assertEqual(len(jobs), 3)

        # Jobs should be ordered by start_time DESC (newest first)
        self.assertEqual(jobs[0].id, job3.id)
        self.assertEqual(jobs[1].id, job2.id)
        self.assertEqual(jobs[2].id, job1.id)

    def test_get_jobs_by_status(self):
        """Test filtering jobs by status."""
        job1 = db_util.start_job("Process1", "Scheduler")
        job2 = db_util.start_job("Process2", "Scheduler")
        job3 = db_util.start_job("Process3", "Scheduler")

        db_util.set_job_status(str(job1.id), JobStatus.DONE)
        db_util.set_job_status(str(job2.id), JobStatus.FAILED)
        # job3 remains RUNNING

        running_jobs = db_util.get_jobs(status=JobStatus.RUNNING)
        self.assertEqual(len(running_jobs), 1)
        self.assertEqual(running_jobs[0].id, job3.id)

        done_jobs = db_util.get_jobs(status=JobStatus.DONE)
        self.assertEqual(len(done_jobs), 1)
        self.assertEqual(done_jobs[0].id, job1.id)

        failed_jobs = db_util.get_jobs(status=JobStatus.FAILED)
        self.assertEqual(len(failed_jobs), 1)
        self.assertEqual(failed_jobs[0].id, job2.id)

    def test_to_row_dict(self):
        """Test the to_row_dict method."""
        job = db_util.start_job("Test Process", "Test Scheduler")

        row_dict = job.to_row_dict()

        self.assertEqual(row_dict["ID"], str(job.id))
        self.assertEqual(row_dict["Process Name"], "Test Process")
        self.assertEqual(row_dict["Scheduler"], "Test Scheduler")
        self.assertRegex(row_dict["Start Time"], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(row_dict["End Time"], "N/A")
        self.assertEqual(row_dict["Status"], "Running")

        # Test with completed job
        db_util.set_job_status(str(job.id), JobStatus.DONE)

        updated_job = db_util.get_job(job.id)
        row_dict = updated_job.to_row_dict()

        self.assertRegex(row_dict["End Time"], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(row_dict["Status"], "Done")


if __name__ == '__main__':
    unittest.main()



"""Module for the Scheduler ORM class"""

from datetime import datetime
from typing import Optional

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.database.base import Base

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class Scheduler(Base):
    """Class containing the ORM model for schedulers"""
    __tablename__ = "Schedulers"

    machine_name: Mapped[str] = mapped_column(String(100), primary_key=True)
    last_update: Mapped[datetime] = mapped_column(onupdate=datetime.now, default=datetime.now)
    latest_trigger: Mapped[Optional[str]] = mapped_column(String(100))
    latest_trigger_time: Mapped[Optional[datetime]]

    def to_row_dict(self) -> dict[str, str]:
        """Convert scheduler to a row dictionary for display in a table."""
        return {
            "Machine Name": self.machine_name,
            "Last Connection": datetime_util.format_datetime(self.last_update),
            "Latest Trigger": self.latest_trigger if self.latest_trigger else "None yet",
            "Latest Trigger Time": datetime_util.format_datetime(self.latest_trigger_time) if self.latest_trigger_time else ""
        }



"""This module contains a single class: ConnectionFrame."""

import os

from nicegui import ui

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.orchestrator import test_helper
from OpenOrchestrator.orchestrator.popups import generic_popups


# pylint: disable-next=too-few-public-methods
class ConnectionFrame():
    """A ui module containing input and buttons for connecting to the database."""
    def __init__(self):
        self.conn_input = ui.input("Connection String").classes("w-4/6")
        self.key_input = ui.input("Encryption Key").classes("w-4/6")
        self._define_validation()
        with ui.row().classes("w-full"):
            self.conn_button = ui.button("Connect", on_click=self._connect)
            self.disconn_button = ui.button("Disconnect", on_click=self._disconnect, )
            self.disconn_button.disable()

        self._initial_connect()
        test_helper.set_automation_ids(self, "connection_frame")

    def _define_validation(self):
        self.conn_input._validation = {"Please enter a connection string": bool}  # pylint: disable=protected-access
        self.key_input._validation = {"Invalid AES key": crypto_util.validate_key}  # pylint: disable=protected-access

    async def _connect(self) -> None:
        """Validate the connection string and encryption key
        and connect to the database.
        """
        if not self.conn_input.validate() & self.key_input.validate():
            ui.notify("Please fill out all fields.", type='warning')
            return

        conn_string = self.conn_input.value
        crypto_key = self.key_input.value

        if db_util.connect(conn_string):
            crypto_util.set_key(crypto_key)
            self._set_state(True)
            ui.notify("Connected!", type='positive', timeout=1000)

            if not os.getenv("ORCHESTRATOR_TEST") and not db_util.check_database_revision():
                await generic_popups.info_popup("This version of Orchestrator doesn't match the version of the connected database. Unexpected errors might occur.")

    def _disconnect(self) -> None:
        db_util.disconnect()
        crypto_util.set_key(None)
        self._set_state(False)
        ui.notify("Disconnected!", type='positive')

    def _set_state(self, connected: bool) -> None:
        if connected:
            self.conn_input.disable()
            self.key_input.disable()
            self.conn_button.disable()
            self.disconn_button.enable()
        else:
            self.conn_input.enable()
            self.key_input.enable()
            self.conn_button.enable()
            self.disconn_button.disable()

    def _initial_connect(self) -> None:
        """Check the environment for a connection string
        and encryption key and connect to the database if both
        are found.
        """
        conn_string = os.environ.get('OpenOrchestratorConnString', None)
        if conn_string:
            self.conn_input.value = conn_string

        crypto_key = os.environ.get('OpenOrchestratorKey', None)
        if crypto_key:
            self.key_input.value = crypto_key

    def new_key(self):
        """Creates a new encryption key and inserts it
        into the key entry.
        """
        key = crypto_util.generate_key()
        self.key_input.value = key.decode()
